import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional, Union

from attrs import converters, define, field, validators

logger = logging.getLogger(__name__)


def _dt_converter(value: str) -> datetime:
    """Convert string to datetime.

    Args:
        value: String with format %d/%m/%Y %H:%M
    Returns:
        Datetime object
    """
    try:
        format = "%d/%m/%Y %H:%M"
        return datetime.strptime(value, format)
    except ValueError:
        raise ValueError("Datetime should be provided as follows: dd/mm/yyyy hh:mm")


def _tablestr_converter(value: Union[str, List[str]]) -> List[str]:
    """Returns a list of table strings, even if a single string is provided

    Args:
        value: Table string or list of table strings
    Returns:
        List of strings
    """
    if type(value) is str:
        return [value]
    else:
        return list(value)


def _validate_forecast_chronology(instance, attribute, value):
    """Validates forecast_start against forecast_end"""
    if value > instance.forecast_end:
        raise ValueError(
            "Forecast end datetime must be greater than or equal to"
            + " forecast start datetime."
        )


def _validate_forecasted_chronology(instance, attribute, value):
    """Validated forecasted_start against forecasted_end"""
    if value > instance.forecasted_end:
        raise ValueError(
            "Forecasted end datetime must be greater than or equal to"
            + " forecasted start datetime."
        )


def _validate_relative_chronology(instance, attribute, value):
    """Validates forecast_start against forecasted_start"""
    if value >= instance.forecasted_start:
        raise ValueError(
            "Forecasted start datetime should be after forecast start datetime."
        )


def _validate_path(instance, attribute, value):
    """Check the path is a directory and creates it if it is not"""
    if not value.is_dir():
        value.mkdir()
        logger.info(f"Created directory at {value.absolute()}")


def _validate_raw_not_processed(instance, attribute, value):
    """Check that `raw_cache` and `processed_cache` are distinct."""
    if instance.processed_cache:
        if value.absolute() == instance.processed_cache.absolute():
            raise ValueError(
                f"{attribute.name} should be distinct from processed_cache"
            )


@define
class Loader:
    """`Loader` validates user inputs and dispatches data fetchers.

    Construct `Loader` using the `Loader.initialise()` constructor. This
    ensures query metadata is constructed approriately.

    Loader:

    - Validates user input data
        - Checks datetime are dd/mm/yyyy HH:MM
        - Checks datetime chronology (e.g. end is after start)
        - Validates `forecast_type`
        - *Validates user-supplied tables against what is available on NEMWeb*
    - Retains query metadata (via constructor class method `initialise`)
    - Can dispatch various Managers and Downloaders

    Attributes:
        forecast_start: Forecasts made at or after this datetime are queried.
        forecast_end: Forecasts made before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: `MTPASA`, `STPASA`, `PDPASA`, `PREDISPATCH` or `P5MIN`.
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        metadata: Metadata dictionary. Constructed by `Loader.initialise()`.
        raw_cache (optional): Path to build or reuse raw cache.
        processed_cache (optional): Path to build or reuse processed cache. Should be
          distinct from raw_cache

    """

    forecast_start: datetime = field(
        converter=_dt_converter,
        validator=[_validate_forecast_chronology, _validate_relative_chronology],
    )
    forecast_end: datetime = field(converter=_dt_converter)
    forecasted_start: datetime = field(
        converter=_dt_converter, validator=[_validate_forecasted_chronology]
    )
    forecasted_end: datetime = field(converter=_dt_converter)
    forecast_type: str = field(
        validator=validators.in_(["MTPASA", "STPASA", "PDPASA", "PREDISPATCH", "P5MIN"])
    )
    tables: List[str] = field(converter=_tablestr_converter)
    metadata: Dict
    raw_cache: Optional[str] = field(
        default=None,
        converter=converters.optional(Path),
        validator=[
            validators.optional(_validate_path),
            validators.optional(_validate_raw_not_processed),
        ],
    )
    processed_cache: Optional[str] = field(
        default=None,
        converter=converters.optional(Path),
        validator=validators.optional(_validate_path),
    )

    @classmethod
    def initialise(
        cls,
        forecast_start: str,
        forecast_end: str,
        forecasted_start: str,
        forecasted_end: str,
        forecast_type: str,
        tables: Union[str, List[str]],
        raw_cache: Optional[str] = None,
        processed_cache: Optional[str] = None,
    ) -> "Loader":
        """Constructor method for Loader. Assembles query metatdata."""
        metadata = {
            "forecast_start": forecast_start,
            "forecast_end": forecast_end,
            "forecasted_start": forecasted_start,
            "forecasted_end": forecasted_end,
            "forecast_type": forecast_type,
            "tables": tables,
        }
        return cls(
            forecast_start=forecast_start,
            forecast_end=forecast_end,
            forecasted_start=forecasted_start,
            forecasted_end=forecasted_end,
            forecast_type=forecast_type,
            tables=tables,
            metadata=metadata,
            raw_cache=raw_cache,
            processed_cache=processed_cache,
        )



# read version from installed package
import logging
import sys
from importlib.metadata import version

from .nemseer import download_raw_data

__version__ = version("nemseer")

logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.basicConfig(
    stream=sys.stdout, level=logging.INFO, format="%(levelname)s: %(message)s"
)

__all__ = [
    "download_raw_data",
]



import logging
from datetime import datetime
from typing import List, Optional

from attrs import define, field
from dateutil import rrule

from .dl_helpers.functions import (
    _construct_sqlloader_forecastdata_url,
    get_historical_forecast_tables,
    get_sqlloader_filesize,
    get_unzipped_csv,
)
from .loader import Loader

logger = logging.getLogger(__name__)


def _enumerate_tables(tables: List[str], table_str: str, range_to: int):
    """Given a table name, populates a list with enumerated table names

    For example, given 'CONSTRAINTSOLUTION' and `range_to`=3, will populate
    `tables` with ['CONSTRAINTSOLUTION1',...,'CONSTRAINTSOLUTION3'].

    Args:
        tables: Table list
        table_str: Table string to enumerate
        range_to: Integer to enumerate to
    Returns:
        `tables` with enumerated `table_str`
    """
    tables.remove(table_str)
    for i in range(1, range_to + 1):
        tables.append(f"{table_str}{i}")
    return tables


def _validate_tables_on_forecast_start(instance, attribute, value):
    """Validates tables for the provided forecast type.

    Checks user-supplied tables against tables available in MMS Historical
    Data SQL Loader for the month and year of forecast_start.
    """
    start_dt = instance.forecast_start
    tables = get_historical_forecast_tables(
        start_dt.year, start_dt.month, instance.forecast_type
    )
    if not set(value).issubset(set(tables)):
        raise ValueError(
            "Table not available from MMS Historical Data SQL Loader"
            + f" (for {start_dt.month}/{start_dt.year}).\n"
            + f"Tables include: {tables}"
        )


@define(kw_only=True)
class ForecastTypeDownloader:
    forecast_start: datetime
    forecast_end: datetime
    forecast_type: str
    tables: List[str] = field(validator=_validate_tables_on_forecast_start)
    raw_cache: Optional[str] = field(default=None)

    @classmethod
    def from_Loader(cls, loader: Loader):
        """Constructor method for ForecastTypeDownloader from Loader."""
        tables = loader.tables
        if "CONSTRAINTSOLUTION" in tables:
            tables = _enumerate_tables(tables, "CONSTRAINTSOLUTION", 4)
        return cls(
            forecast_start=loader.forecast_start,
            forecast_end=loader.forecast_end,
            forecast_type=loader.forecast_type,
            tables=tables,
            raw_cache=loader.raw_cache,
        )

    def download_zip(self):
        """Downloads zip files given query loaded into ForecastTypeDownloader

        .. todo:: Extend to converting to parquet
        """
        intervening_dates = rrule.rrule(
            rrule.MONTHLY, dtstart=self.forecast_start, until=self.forecast_end
        )
        for table in self.tables:
            for date in intervening_dates:
                (year, month) = (date.year, date.month)
                url = _construct_sqlloader_forecastdata_url(
                    year, month, self.forecast_type, table
                )
                size = get_sqlloader_filesize(year, month, self.forecast_type, table)
                logger.info(
                    f"Downloading {table} for {month}/{year}: {size} MB (zipped)"
                )
                get_unzipped_csv(url, self.raw_cache)



def validate_P5MIN_inputs():
    """
    .. todo:: Create `P5MIN` validator
    """
    return None


def validate_PREDISPATCH_inputs():
    """
    .. todo:: Create `PREDISPATCH` validator
    """
    return None


def validate_PDPASA_inputs():
    """
    .. todo:: Create `PDPASA` validator
    """
    return None


def validate_STPASA_inputs():
    """
    .. todo:: Create `STPASA` validator
    """
    return None


def validate_MTPASA_inputs():
    """
    .. todo:: Create `MTPASA` validator
    .. todo:: Handle MTPASA DUID Availability
    """
    return None



import io
import logging
from itertools import cycle
from pathlib import Path
from re import match, search
from typing import Dict, Generator, List
from zipfile import ZipFile

import requests
from bs4 import BeautifulSoup

from .data import MMSDM_ARCHIVE_URL, USER_AGENTS

logger = logging.getLogger(__name__)


def _build_useragent_generator(n: int) -> Generator:
    """Generator function that cycles through user agents for GET requests.

    Generator function that cycles through user agents to yield n user agents
    in total. Doing so minimises 403 Forbidden errors when scraping.

    Args:
        n: Number of user agents, i.e. number of GET requests.
    Yields:
        useragent: A user agent.
    """
    inf_agents = cycle(USER_AGENTS)
    n_iterator = range(n)
    for _, useragent in zip(n_iterator, inf_agents):
        yield useragent


def _build_nemweb_get_header(useragent: str) -> Dict[str, str]:
    """Builds request header for GET requests from NEMWeb

    Args:
        useragent: User-Agent string to use
    Returns:
        Dict that can be used as a request header
    """
    header = {
        "Host": "www.nemweb.com.au",
        "User-Agent": useragent,
        "Accept": (
            "text/html,application/xhtml+xml,application/xml;"
            + "q=0.9,image/avif,image/webp,*/*;q=0.8"
        ),
        "Accept-Language": "en-US,en;q=0.5",
        "Accept-Encoding": "gzip, deflate",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1",
    }
    return header


def _request_content(
    url: str, useragent: str, additional_header: Dict = {}
) -> requests.Response:
    """Initiates a GET request with header information.

    Args:
        url: URL for GET request.
        useragent: User-Agent to use in header.
        additional_header: Empty dictionary as default. Can be used to add
            additional header information to GET request.
    Returns:
        requests Response object.
    """
    header = _build_nemweb_get_header(useragent)
    if additional_header:
        header.update(additional_header)
    r = requests.get(url, headers=header)
    return r


def _rerequest_to_obtain_soup(
    url: str, useragent: str, additional_header: Dict = {}
) -> BeautifulSoup:
    """Continually launches requests until a 200 (OK) code is returned.

    Args:
        url: URL for GET request.
        useragent: User-Agent to use in header.
        additional_header: Empty dictionary as default. Can be used to add
            additional header information to GET request.

    Returns:
        BeautifulSoup object with parsed HTML.

    """
    r = _request_content(url, useragent, additional_header=additional_header)
    while (ok := r.status_code == requests.status_codes.codes["OK"]) < 1:
        r = _request_content(url, useragent, additional_header=additional_header)
        if r.status_code == requests.status_codes.codes["OK"]:
            ok += 1
    soup = BeautifulSoup(r.content, "html.parser")
    return soup


def _get_years_and_months() -> Dict[int, List[int]]:
    """Checks months with data available, for each year from scraped link.

    Returns:
        Months mapped to each year. Data is available for each of these months.
    """

    def _get_months(url: str, useragent: str) -> List[int]:
        """Pull months from scraped links with YYYY-MM date format

        Args:
            url: url for GET request.
            header: useragent to pass to GET request.
        Returns:
            List of unique months (as integers).
        """
        referer_header = {"Referer": MMSDM_ARCHIVE_URL}
        soup = _rerequest_to_obtain_soup(
            url, useragent, additional_header=referer_header
        )
        months = []
        for link in soup.find_all("a"):
            url = link.get("href")
            findmonth = match(r".*[0-9]{4}_([0-9]{2})", url)
            if not findmonth:
                continue
            else:
                month = findmonth.group(1)
                months.append(int(month))
        unique = list(set(months))
        return unique

    useragent = next(_build_useragent_generator(1))
    soup = _rerequest_to_obtain_soup(MMSDM_ARCHIVE_URL, useragent)
    links = soup.find_all("a")
    nlinks = len(links)
    yearmonths = {}
    for useragent, link in zip(_build_useragent_generator(nlinks), links):
        url = link.get("href")
        findyear = match(r".*([0-9]{4}).*", url)
        if not findyear:
            continue
        else:
            year = int(findyear.group(1))
            months = _get_months(MMSDM_ARCHIVE_URL + f"{year}/", useragent)
            yearmonths[year] = months
    return yearmonths


def _construct_sqlloader_yearmonth_url(year: int, month: int) -> str:
    """Constructs MMSDM Historical Data SQLLoader URL for a given year and month

    Args:
        year: Year
        month: Month
    Returns:
        Constructed URL as a string.
    """
    url = (
        MMSDM_ARCHIVE_URL
        + f"{year}/MMSDM_{year}_"
        + f'{str(month).rjust(2, "0")}/MMSDM_Historical_Data_SQLLoader/'
        + "DATA/"
    )
    return url


def _construct_sqlloader_forecastdata_url(
    year: int, month: int, forecast_type: str, table: str
) -> str:
    """Constructs URL that points to a NEMWeb zip file

    Handles exceptions to naming rules for `PREDISPATCH`

    Args:
        year: Year
        month: Month
        forecast_type: `P5MIN`, `PREDISPATCH`, `PDPASA`, `STPASA` or `MTPASA`
        table: The name of the table required
    Returns:
        URL pointing to forecast data zipfile on NEMWeb
    """
    base_url = _construct_sqlloader_yearmonth_url(year, month)
    (stryear, strmonth) = (str(year), str(month).rjust(2, "0"))
    if forecast_type == "PREDISPATCH" and forecast_type != "MNSPBIDTRK":
        prefix = f"PUBLIC_DVD_{forecast_type}{table}"
    else:
        prefix = f"PUBLIC_DVD_{forecast_type}_{table}"
    url = base_url + prefix + f"_{stryear}{strmonth}010000.zip"
    return url


def get_sqlloader_filesize(
    year: int, month: int, forecast_type: str, table: str
) -> int:
    """For a particular table file on NEMWeb, scrapes and returns file size in MB

    NEMWeb has file size in a column preceding the link to the file. This functions
    scrapes and returns a megabyte filesize (NEMWeb file size is in bytes).

    Args:
        year: Year
        month: Month
        forecast_type: `P5MIN`, `PREDISPATCH`, `PDPASA`, `STPASA` or `MTPASA`
        table: The name of the table required
    Returns:
        File size in megabytes, rounded to nearest MB
    """
    parent_url = _construct_sqlloader_yearmonth_url(year, month)
    data_url = _construct_sqlloader_forecastdata_url(year, month, forecast_type, table)
    data_table = data_url.lstrip(parent_url)
    useragent = next(_build_useragent_generator(1))
    soup = _rerequest_to_obtain_soup(parent_url, useragent)
    if not (size_and_file := search(f"([0-9]*) {data_table}", soup.get_text())):
        raise ValueError(f" Cannot find file size for {data_table}")
    else:
        size = size_and_file.group(1)
        size = round(float(size) / (1024**2))
    return size


def get_historical_forecast_tables(
    year: int, month: int, forecast_type: str
) -> List[str]:
    """For a month & year, get available tables for a particular forecast type.

    Handling of special cases:
      - Removes numbering from enumerated tables for `P5MIN` (`CONSTRAINTSOLUTION(x)`)

    Args:
        year: Year
        month: Month
        forecast_type: AEMO forecast types (`P5MIN`, `PREDISPATCH`, `STPASA`, `MTPASA`)
    Returns:
        List of tables associated with that forecast type for that period
    """

    def _unique_tablename_from_enumerated(links: List[str]) -> List[str]:
        """Handles enumerated table names

        Handles P5MIN tables that are spread across multiple files,
         e.g. CONSTRAINTSOLUTION1, CONSTRAINTSOLUTION2, etc.

        Args:
            links: A list of links
        Returns:
            List of unique tables, with enumerated links/tables collapsed
        """
        table_capture = f".*/PUBLIC_DVD_{forecast_type}([A-Z_]*)[0-9]?_[0-9]*.zip"
        tables = []
        for link in links:
            if mo := match(table_capture, link):
                tables.append(mo.group(1).lstrip("_"))
        return list(set(tables))

    url = _construct_sqlloader_yearmonth_url(year, month)
    soup = _rerequest_to_obtain_soup(url, next(_build_useragent_generator(1)))
    links = [link.get("href") for link in soup.find_all("a")]
    tables = _unique_tablename_from_enumerated(links)
    return tables


def get_unzipped_csv(url: str, raw_cache: Path) -> None:
    """Downloads unzipped (single) csv file from `url` to `raw_cache`

    Validates that the zip contains a single file that has a similar name to the zip

    Args:
        url: URL of zip
        raw_cache: Path to save zip
    Returns:
        None. Extracts csvs to `raw_cache`.
    """
    r = _request_content(url, next(_build_useragent_generator(1)))
    z = ZipFile(io.BytesIO(r.content))
    if (
        len(csvfn := z.namelist()) == 1
        and (zfn := match(".*/DATA/(.*).zip", url))
        and (fn := match("(.*).[cC][sS][vV]", csvfn.pop()))
        and (fn.group(1) == zfn.group(1))
    ):
        z.extractall(raw_cache)
    else:
        raise ValueError(f"Unexpected contents in zipfile from {url}")



from typing import List, Union

from .downloader import ForecastTypeDownloader
from .loader import Loader


def download_raw_data(
    forecast_start: str,
    forecast_end: str,
    forecasted_start: str,
    forecasted_end: str,
    forecast_type: str,
    tables: Union[str, List[str]],
    raw_cache: str,
) -> None:
    """Downloads raw forecast data from NEMWeb

    Arguments:
        forecast_start: Forecasts made at or after this datetime are queried.
        forecast_end: Forecasts made before or at this datetime are queried.
        forecasted_start: Forecasts pertaining to times at or after this
            datetime are retained.
        forecasted_end: Forecasts pertaining to times before or at this
            datetime are retaned.
        forecast_type: `MTPASA`, `STPASA`, `PDPASA`, `PREDISPATCH` or `P5MIN`.
        tables: Table or tables required. A single table can be supplied as
            a string. Multiple tables can be supplied as a list of strings.
        raw_cache: Path to download files
    """
    loader = Loader.initialise(
        forecast_start=forecast_start,
        forecast_end=forecast_end,
        forecasted_start=forecasted_start,
        forecasted_end=forecasted_end,
        forecast_type=forecast_type,
        tables=tables,
        raw_cache=raw_cache,
    )
    downloader = ForecastTypeDownloader.from_Loader(loader)
    downloader.download_zip()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nemseer', 'nemseer.dl_helpers']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21,<22',
 'beautifulsoup4>=4,<5',
 'dask>=2022.7.1,<2023.0.0',
 'netCDF4>=1.6.0,<2.0.0',
 'numpy>=1.23.0,<2.0.0',
 'packaging>=21.3,<22.0',
 'pandas>=1.2,<2.0',
 'pyarrow>=8.0.0,<9.0.0',
 'requests>=2,<3',
 'tqdm>=4.64.0,<5.0.0',
 'xarray>=2022,<2023']

setup_kwargs = {
    'name': 'nemseer',
    'version': '0.1.0',
    'description': 'A package for downloading and handling forecasts for the National Electricity Market (NEM) from the Australian Energy Market Operator (AEMO).',
    'long_description': "# nemseer\n\n[![Documentation Status](https://readthedocs.org/projects/nemseer/badge/?version=latest)](https://nemseer.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/UNSW-CEEM/NEMSEER/branch/master/graph/badge.svg?token=BO69YSQIGI)](https://codecov.io/gh/UNSW-CEEM/NEMSEER)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nA package for downloading and handling forecasts for the National Electricity Market (NEM) from the Australian Energy Market Operator (AEMO).\n\n## Work in Progress\n\nThis package is a work in progress. For a high-level overview of development, check out the [roadmap](ROADMAP.md).\n\n## Installation\n\n- TODO\n\n## Usage\n\n- TODO\n\n## Contributing\n\nInterested in contributing? Check out the [contributing guidelines](CONTRIBUTING.md), which also includes steps to install `nemseer` for development.\n\nPlease note that this project is released with a [Code of Conduct](CONDUCT.md). By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`nemseer` was created by Abhijith Prakash. It is licensed under the terms of the [BSD 3-Clause license](LICENSE).\n\n## Credits\n\n`nemseer` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n\n`nemseer` borrows functionality from [`NEMOSIS`](https://github.com/UNSW-CEEM/NEMOSIS), a package for extracting historical *actual* market data from AEMO's NemWeb.\n",
    'author': 'Abhijith Prakash',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
